# Copyright 2020 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging

from flask import send_file, abort, make_response

from .models import get_setting

log = logging.getLogger(__name__)


KKAGENT_DIR = os.environ.get('KKAGENT_DIR', '')


INSTALL_SCRIPT_SH = '''#!/bin/bash
set -x

KRAKEN_URL="{url}"

# check sudo
sudo -n true
if [ $? -ne 0 ]; then
    echo "error: sudo with no password is required"
    exit 1
fi

# check python3
python3 --version
if [ $? -ne 0 ]; then
    echo "error: missing python3, it is required to run Kraken agent"
    exit 1
fi

# check curl and wget
DL_TOOL=
which curl
if [ $? -ne 0 ]; then
    which wget
    if [ $? -ne 0 ]; then
        echo "error: missing curl and wget, install any of them"
        exit 1
    else
        DL_TOOL=wget
    fi
else
    DL_TOOL=curl
fi

set -e

if [ "$DL_TOOL" == "wget" ]; then
    wget -O /tmp/kkagent ${KRAKEN_URL}/bk/install/agent
else
    curl -o /tmp/kkagent ${KRAKEN_URL}/bk/install/agent
fi
chmod a+x /tmp/kkagent

export KRAKEN_CLICKHOUSE_ADDR="{clickhouse-addr}"

/tmp/kkagent install -s ${KRAKEN_URL}
rm -f /tmp/kkagent
echo 'Kraken Agent installed'
'''

# powershell Invoke-WebRequest -Uri 'http://192.168.68.113:8080/install/kraken-agent-install.bat' -OutFile kraken-agent-install.bat

INSTALL_SCRIPT_BAT = '''
<# :
@echo off
setlocal
set "POWERSHELL_BAT_ARGS=%*"
if defined POWERSHELL_BAT_ARGS set "POWERSHELL_BAT_ARGS=%POWERSHELL_BAT_ARGS:"=\"%"
endlocal & powershell -NoLogo -NoProfile -Command "$input | &{ [ScriptBlock]::Create( ( Get-Content \"%~f0\" ) -join [char]10 ).Invoke( @( &{ $args } %POWERSHELL_BAT_ARGS% ) ) }"
goto :EOF
#>

Write-Host 'Downloading Kraken Agent'

#    Set-PSDebug -Trace 2
$global:ProgressPreference = 'SilentlyContinue'

$PythonPath = Get-Command 'python.exe' | Select -ExpandProperty 'path'
#if ($PythonPath -eq "") {
#    throw "Cannot find path to Python. Please, install it."
#}
Write-Host "Path to Python: $PythonPath"

$KrakenURL = '{url}'
$KKAgentURL = "$KrakenURL/bk/install/agent"

$KKAgentPath = Join-Path $env:Temp 'kkagent'

if (Test-Path -Path $KKAgentPath -PathType Leaf) {
    Remove-Item $KKAgentPath
}

Invoke-WebRequest $KKAgentURL -OutFile $KKAgentPath
Write-Host "Downloaded Kraken Agent to $KKAgentPath"

$env:KRAKEN_CLICKHOUSE_ADDR = '{clickhouse-addr}'

$Cmd = "$KKAgentPath install -s $KrakenURL"
Write-Host "Invoke cmd: $PythonPath $Cmd"
$p = Start-Process -PassThru -NoNewWindow -Wait -FilePath $PythonPath -ArgumentList $Cmd
if ($p.ExitCode -gt 0) {
    throw "Installing Kraken Agent failed"
}

Remove-Item $KKAgentPath

Write-Host 'Kraken Agent installed'
'''


def serve_agent_blob(blob):
    if blob not in ['agent', 'tool', 'kraken-agent-install.sh', 'kraken-agent-install.bat', 'nssm-2.24.zip']:
        abort(404)

    if blob.startswith('kraken-agent-install.'):
        # get and check server url
        url = get_setting('general', 'server_url')
        if not url:
            abort(500, 'Cannot get server URL and put it in Kraken agent install script. ' +
                  'The URL should be set on Kraken Settings page first.')

        # get and check clickhouse addr
        clickhouse_addr = get_setting('general', 'clickhouse_addr')
        if not clickhouse_addr:
            abort(500, 'Cannot get ClickHouse address and put it in Kraken agent install script. ' +
                  'The ClickHouse address should be set on Kraken Settings page first.')

        # patch install script with url and addresses

        if blob.endswith('.bat'):
            script = INSTALL_SCRIPT_BAT
        else:
            script = INSTALL_SCRIPT_SH
        script = script.replace('{url}', url)
        script = script.replace('{clickhouse-addr}', clickhouse_addr)

        resp = make_response(script)

        # add a filename
        if blob.endswith('.bat'):
            resp.headers.set("Content-Type", "application/x-bat")
        else:
            resp.headers.set("Content-Type", "application/x-sh")
        resp.headers.set("Content-Disposition", "attachment", filename=blob)
        return resp

    if blob in ['agent', 'tool']:
        blob = 'kk' + blob

    p = os.path.join(KKAGENT_DIR, blob)
    return send_file(p)



# Copyright 2020-2023 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import hmac
import hashlib
import logging
import datetime

from flask import Blueprint, request, abort
import dateutil.parser

from .models import Project
from .bg import jobs as bg_jobs
from . import kkrq

log = logging.getLogger(__name__)


### GITHUB #############

def handle_github_webhook(project_id):
    payload = request.get_data()
    log.info('GITHUB for project_id:%s, payload: %s', project_id, payload)
    event = request.headers.get('X-GitHub-Event')
    signature = request.headers.get("X-Hub-Signature")
    log.info('GITHUB event:%s, signature: %s', event, signature)
    return _handle_github_webhook(project_id, payload, event, signature)


def _handle_github_webhook(project_id, payload, event, signature):
    # check event
    if event is None:
        log.warning('missing github event type in request header')
        abort(400, "missing github event type in request header")
    if event not in ['push', 'pull_request']:
        msg = 'unsupported event'
        log.info(msg)
        return msg, 204

    # check project
    project = Project.query.filter_by(id=project_id).one_or_none()
    if project is None:
        log.warning('cannot find project %s', project_id)
        abort(400, "Invalid project id")

    if not project.webhooks or not project.webhooks.get('github_enabled', False):
        log.info('webhooks from github disabled')
        abort(400, "webhooks from github disabled")

    # check secret
    my_secret = project.webhooks.get('github_secret', None)
    if my_secret:
        my_secret = bytes(my_secret, 'ascii')
    if my_secret is not None:
        if signature is None:
            log.warning('missing signature in request header')
            abort(400, "missing signature in request header")
        github_digest_parts = signature.split("=", 1)
        my_digest = hmac.new(my_secret, payload, hashlib.sha1).hexdigest()

        if len(github_digest_parts) < 2 or github_digest_parts[0] != "sha1" or not hmac.compare_digest(github_digest_parts[1], my_digest):
            log.warning('bad signature %s vs %s', signature, my_digest)
            abort(400, "Invalid signature")

    req = json.loads(payload)

    # trigger running the project flow via rq
    if event == 'push':
        trigger_data = dict(trigger='github-' + event,
                            ref=req['ref'],
                            before=req['before'],
                            after=req['after'],
                            repo=req['repository']['clone_url'],
                            pusher=req['pusher'],
                            commits=req['commits'])
    elif event == 'pull_request':
        if req['action'] not in ['opened', 'synchronize']:
            msg = 'unsupported action %s' % req['action']
            log.info(msg)
            return msg, 204

        if req['action'] == 'opened' and req['pull_request']['commits'] == 0:
            msg = 'pull request with no commits, dropped'
            log.info(msg)
            return msg, 204

        if 'before' in req:
            before = req['before']
        else:
            before = req['pull_request']['base']['sha']

        if 'after' in req:
            after = req['after']
        else:
            after = req['pull_request']['head']['sha']

        trigger_data = dict(trigger='github-' + event,
                            action=req['action'],
                            pull_request=req['pull_request'],
                            before=before,
                            after=after,
                            repo=req['repository']['clone_url'],
                            sender=req['sender'])
    kkrq.enq(bg_jobs.trigger_flow, project.id, trigger_data)
    return "", 204


### GITEA #############

def handle_gitea_webhook(project_id):
    payload = request.get_data()
    log.info('GITEA for project_id:%s, payload: %s', project_id, payload)
    event = request.headers.get('X-Gitea-Event')
    signature = request.headers.get("X-Hub-Signature")
    log.info('GITEA event:%s, signature: %s', event, signature)
    return _handle_gitea_webhook(project_id, payload, event, signature)


def _handle_gitea_webhook(project_id, payload, event, signature):
    # check event
    if event is None:
        log.warning('missing gitea event type in request header')
        abort(400, "missing gitea event type in request header")
    if event not in ['push', 'pull_request']:
        msg = 'unsupported event'
        log.info(msg)
        return msg, 204

    # check project
    project = Project.query.filter_by(id=project_id).one_or_none()
    if project is None:
        log.warning('cannot find project %s', project_id)
        abort(400, "Invalid project id")

    if not project.webhooks or not project.webhooks.get('gitea_enabled', False):
        log.info('webhooks from gitea disabled')
        abort(400, "webhooks from gitea disabled")

    # check secret
    my_secret = project.webhooks.get('gitea_secret', None)
    if my_secret:
        my_secret = bytes(my_secret, 'ascii')
    if my_secret is None:
        msg = 'secret not configured for gitea webhook'
        log.warning(msg)
        return msg, 204
    if signature is None:
        log.warning('missing signature in request header')
        abort(400, "missing signature in request header")
    digest_parts = signature.split("=", 1)
    my_digest = hmac.new(my_secret, payload, hashlib.sha1).hexdigest()

    if len(digest_parts) < 2 or digest_parts[0] != "sha1" or not hmac.compare_digest(digest_parts[1], my_digest):
        log.warning('bad signature %s vs %s', signature, my_digest)
        abort(400, "Invalid signature")

    req = json.loads(payload)

    # trigger running the project flow via rq
    if event == 'push':
        trigger_data = dict(trigger='gitea-' + event,
                            ref=req['ref'],
                            before=req['before'],
                            after=req['after'],
                            repo=req['repository']['clone_url'],
                            pusher=req['pusher'],
                            commits=req['commits'])
    elif event == 'pull_request':
        if req['action'] not in ['opened', 'synchronize', 'synchronized']:
            msg = 'unsupported action %s' % req['action']
            log.info(msg)
            return msg, 204

        if 'before' in req and 'after' in req:
            before = req['before']
            after = req['after']
        else:
            before = req['pull_request']['base']['sha']
            after = req['pull_request']['head']['sha']

        if after == before:
            msg = 'pull request with no commits, dropped'
            log.info(msg)
            return msg, 204

        trigger_data = dict(trigger='gitea-' + event,
                            action=req['action'],
                            pull_request=req['pull_request'],
                            before=before,
                            after=after,
                            repo=req['repository']['clone_url'],
                            sender=req['sender'])
    kkrq.enq(bg_jobs.trigger_flow, project.id, trigger_data)

    return "", 204


### GITLAB #############

def handle_gitlab_webhook(project_id):
    payload = request.get_data()
    log.info('GITLAB for project_id:%s, payload: %s', project_id, payload)
    event = request.headers.get('X-Gitlab-Event')
    token = request.headers.get('X-Gitlab-Token')
    log.info('GITLAB event:%s', event)
    return _handle_gitlab_webhook(project_id, payload, event, token)


def _handle_gitlab_webhook(project_id, payload, event, token):
    # check event
    if event is None:
        log.warning('missing gitlab event type in request header')
        abort(400, "missing gitlab event type in request header")
    if event not in ['Push Hook', 'Merge Request Hook']:
        msg = 'unsupported event'
        log.info(msg)
        return msg, 204

    # check project
    project = Project.query.filter_by(id=project_id).one_or_none()
    if project is None:
        log.warning('cannot find project %s', project_id)
        abort(400, "Invalid project id")

    if not project.webhooks or not project.webhooks.get('gitlab_enabled', False):
        log.info('webhooks from gitlab disabled')
        abort(400, "webhooks from gitlab disabled")

    # check secret
    my_secret = project.webhooks.get('gitlab_secret', None)
    if my_secret is None:
        log.warning('secret not configured for gitea webhook')
        return "", 204
    if token is None:
        log.warning('missing token in request header')
        abort(400, "missing token in request header")

    if token != my_secret:
        log.warning('bad token %s vs %s', token, my_secret)
        abort(400, "Invalid token")

    req = json.loads(payload)

    # trigger running the project flow via rq
    if event == 'Push Hook':
        trigger_data = dict(trigger='gitlab-' + event,
                            ref=req['ref'],
                            before=req['before'],
                            after=req['after'],
                            repo=req['repository']['git_http_url'],
                            pusher=dict(full_name=req['user_name'],
                                        username=req['user_username'],
                                        email=req['user_email']),
                            commits=req['commits'])
    elif event == 'Merge Request Hook':
        obj = req['object_attributes']
        action = obj['action']
        if action not in ['open', 'update']:
            msg = 'unsupported action %s' % action
            log.info(msg)
            return msg, 204

        if 'oldrev' in obj:
            before = obj['oldrev']
        else:
            before = ''

        after = obj['last_commit']['id']

        if after == before:
            msg = 'merge request with no commits, dropped'
            log.info(msg)
            return msg, 204

        # get base url
        base_url = req['project']['web_url']
        base_url = base_url.rsplit('/', 2)[0]

        trigger_data = dict(trigger='gitlab-' + event,
                            action=action,
                            pull_request=dict(head=dict(ref=obj['source_branch']),
                                              base=dict(ref=obj['target_branch'],
                                                        sha=before),
                                              user=dict(login=req['user']['username'],
                                                        html_url='%s/%s' % (base_url, req['user']['username'])),
                                              html_url=obj['url'],
                                              number=obj['id'],
                                              updated_at=dateutil.parser.parse(obj['updated_at']).isoformat(),
                                              title=obj['title']),
                            before=before,
                            after=after,
                            repo=req['project']['git_http_url'],
                            sender=req['user'])
    kkrq.enq(bg_jobs.trigger_flow, project.id, trigger_data)

    return "", 204


### RADICLE #############

def handle_radicle_webhook(project_id):
    payload = request.get_data()
    log.info('RADICLE for project_id:%s, payload: %s', project_id, payload)
    event = request.headers.get('X-Radicle-Event-Type')
    # event_key = request.headers.get('X-Event-Key')
    # hook_uuid = request.headers.get('X-Hook-UUID')
    # request_uuid = request.headers.get('X-Request-UUID')
    signature = request.headers.get('X-Hub-Signature-256')
    log.info('RADICLE event:%s, signature: %s', event, signature)
    return _handle_radicle_webhook(project_id, payload, event, signature)


def _handle_radicle_webhook(project_id, payload, event, signature):
    # check event
    if event is None:
        log.warning('missing radicle event type in request header')
        abort(400, "missing radicle event type in request header")
    if event not in ['push', 'patch']:
        msg = 'unsupported event'
        log.info(msg)
        return msg, 204

    # check project
    project = Project.query.filter_by(id=project_id).one_or_none()
    if project is None:
        log.warning('cannot find project %s', project_id)
        abort(400, "Invalid project id")

    if not project.webhooks or not project.webhooks.get('radicle_enabled', False):
        log.info('webhooks from radicle disabled')
        abort(400, "webhooks from radicle disabled")

    # check secret
    my_secret = project.webhooks.get('radicle_secret', None)
    if my_secret:
        my_secret = bytes(my_secret, 'ascii')
    if my_secret is not None:
        if signature is None:
            log.warning('missing signature in request header')
            abort(400, "missing signature in request header")
        digest_parts = signature.split("=", 1)
        my_digest = hmac.new(my_secret, payload, hashlib.sha256).hexdigest()

        if len(digest_parts) < 2 or digest_parts[0] != "sha256" or not hmac.compare_digest(digest_parts[1], my_digest):
            log.warning('bad signature %s vs %s', signature, my_digest)
            abort(400, "Invalid signature")

    req = json.loads(payload)

    # trigger running the project flow via rq
    if event == 'push':
        trigger_data = dict(trigger='radicle-' + event,
                            # ref=req['ref'], TODO
                            ref="refs/heads/" + req['branch'],
                            before=req['before'],
                            after=req['after'],
                            repo=req['repository']['http_url'],
                            pusher=dict(#full_name=req['author'],
                                        username=req['author']['alias'],
                                        #email=req['author']
                            ),
                            commits=req['commits'])
    elif event == 'patch':
        action = req['action']
        if action not in ['created', 'updated']:
            msg = 'unsupported action %s' % action
            log.info(msg)
            return msg, 204

        patch = req['patch']

        before = patch['before']
        after = patch['after']

        if after == before:
            msg = 'patch with no commits, dropped'
            log.info(msg)
            return msg, 204

        trigger_data = dict(trigger='radicle-' + event,
                            action=action,
                            pull_request=dict(#head=dict(ref=obj['source_branch']),
                                              #base=dict(ref=obj['target_branch'],
                                              #          sha=before),
                                              base=dict(ref=req['repository']['default_branch']),  # TODO
                                              user=dict(login=patch['author']['alias']),
                                                        #html_url='%s/%s' % (base_url, req['user']['username'])),
                                              html_url=patch['url'],
                                              number=patch['id'],
                                              #updated_at=dateutil.parser.parse(obj['updated_at']).isoformat(),
                                              title=patch['title'],
                                              commits=len(patch['revisions']),
                            ),
                            before=before,
                            after=after,
                            repo=req['repository']['http_url'],
                            sender=dict(#full_name=req['author'],
                                        login=patch['author']['alias'],
                                        #email=req['author']
                            ))
    kkrq.enq(bg_jobs.trigger_flow, project.id, trigger_data)

    return "", 204


### BLUEPRINT #############

def create_blueprint():
    bp = Blueprint('webhooks', __name__)

    bp.add_url_rule('/<int:project_id>/github', view_func=handle_github_webhook, methods=['POST'])
    bp.add_url_rule('/<int:project_id>/gitea', view_func=handle_gitea_webhook, methods=['POST'])
    bp.add_url_rule('/<int:project_id>/gitlab', view_func=handle_gitlab_webhook, methods=['POST'])
    bp.add_url_rule('/<int:project_id>/radicle', view_func=handle_radicle_webhook, methods=['POST'])

    return bp



"""fields_raw not nullable

Revision ID: 8e51ac616b91
Revises: 68293032700d
Create Date: 2023-07-28 07:30:34.147891

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '8e51ac616b91'
down_revision = '68293032700d'
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()
    conn.execute("UPDATE steps SET fields_raw = '{}' WHERE fields_raw IS NULL")

    op.alter_column('steps', 'fields_raw',
               existing_type=postgresql.JSONB(astext_type=sa.Text()),
               nullable=False)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('steps', 'fields_raw',
               existing_type=postgresql.JSONB(astext_type=sa.Text()),
               nullable=True)
    # ### end Alembic commands ###



"""added fields_raw to step

Revision ID: 9e57e0f30a5f
Revises: b510fae321ba
Create Date: 2023-06-30 14:28:08.251557

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '9e57e0f30a5f'
down_revision = 'b510fae321ba'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('steps', sa.Column('fields_raw', postgresql.JSONB(astext_type=sa.Text()), nullable=True))


def downgrade():
    op.drop_column('steps', 'fields_raw')



"""added seq to flow and run

Revision ID: b510fae321ba
Revises: bdcdbb4b5b3f
Create Date: 2023-03-19 13:45:12.691045

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'b510fae321ba'
down_revision = 'bdcdbb4b5b3f'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('flows', sa.Column('seq', postgresql.JSONB(astext_type=sa.Text()), nullable=True))
    op.add_column('runs', sa.Column('seq', postgresql.JSONB(astext_type=sa.Text()), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('runs', 'seq')
    op.drop_column('flows', 'seq')
    # ### end Alembic commands ###



# Copyright 2023 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging

import jq
import jsonpatch
from sqlalchemy.orm.attributes import flag_modified


from .models import db


log = logging.getLogger(__name__)


def handle_data(job, step, data):
    operation = step.fields.get('operation', 'set')
    scope = step.fields.get('scope', 'flow')
    ptr = step.fields.get('json_pointer', '/')
    value = step.fields.get('value', None)

    if operation not in ['set', 'jq', 'jsonpatch']:
        raise Exception('incorrect operation %s, should be one of set, jq, jsonpatch' % operation)

    if value:
        data = value

    log.info('handle_data %s %s %s', job, step, data)

    # get proper db entity and field according indicated scope
    if scope == 'project':
        entity = job.run.flow.branch.project
        data_field_name = 'user_data'
    elif scope == 'branch':
        entity = job.run.flow.branch
        data_field_name = 'user_data'
    elif scope == 'branch-ci':
        entity = job.run.flow.branch
        data_field_name = 'user_data_ci'
    elif scope == 'branch-dev':
        entity = job.run.flow.branch
        data_field_name = 'user_data_dev'
    elif scope in ['flow', '']:
        entity = job.run.flow
        data_field_name = 'user_data'
    else:
        log.warning('incorrect scope %s, should be one of project, branch, branch-ci, branch-dev, flow', scope)
        return

    # get current user data from database
    user_data = getattr(entity, data_field_name)

    # prepare data provided by user according to specified operation
    if operation == 'set':
        try:
            data = json.loads(data)
        except Exception:
            log.exception('wrong json data: %s', data)
            return
    elif operation == 'jq':
        try:
            q = jq.compile(data)
        except Exception:
            log.exception('wrong jq expression: %s', data)
            return
    elif operation == 'jsonpatch':
        try:
            patch = jsonpatch.JsonPatch.from_string(data)
        except Exception:
            log.exception('wrong jsonpatch expression: %s', data)
            return

    ptr = ptr.strip('/')
    if ptr:
        # replace user_data under position indicated by json pointer
        parts = ptr.split('/')
        user_data = getattr(entity, data_field_name)
        for idx, p in enumerate(parts):
            if p.isdigit():
                p = int(p)
            if idx != len(parts) - 1:
                user_data = user_data[p]
            else:
                if operation == 'set':
                    user_data[p] = data
                elif operation == 'jq':
                    user_data[p] = q.input(user_data[p]).first()
                elif operation == 'jsonpatch':
                    user_data[p] = patch.apply(user_data[p])
    else:
        # set new data on root level
        if operation == 'set':
            new_user_data = data
        elif operation == 'jq':
            new_user_data = q.input(user_data).first()
        elif operation == 'jsonpatch':
            new_user_data = patch.apply(user_data)

        setattr(entity, data_field_name, new_user_data)

    flag_modified(entity, data_field_name)
    db.session.commit()
